/*-
 * %%----------------------------------------------------------------------------------------------
 * Solidify Framework - Solidify Frontend - download.service.ts
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2023 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

import {
  HttpClient,
  HttpHeaders,
  HttpStatusCode,
} from "@angular/common/http";
import {
  Inject,
  Injectable,
} from "@angular/core";
import {saveAs} from "file-saver";
import {Observable} from "rxjs";
import {
  catchError,
  take,
  tap,
} from "rxjs/operators";
import {LabelTranslateInterface} from "../../label-translate-interface.model";
import {OauthHelper} from "../../oauth2/oauth.helper";
import {SOLIDIFY_CONSTANTS} from "../constants";
import {CookieType} from "../cookie-consent/enums/cookie-type.enum";
import {CookieConsentService} from "../cookie-consent/services/cookie-consent.service";
import {CookieConsentUtil} from "../cookie-consent/utils/cookie-consent.util";
import {ApiActionNamePartialEnum} from "../enums/partial/api-action-name-partial.enum";
import {CookiePartialEnum} from "../enums/partial/cookie-partial.enum";
import {DefaultSolidifyEnvironment} from "../environments/environment.solidify-defaults";
import {SolidifyError} from "../errors/solidify.error";
import {ApiService} from "../http/api.service";
import {ENVIRONMENT} from "../injection-tokens/environment.injection-token";
import {LABEL_TRANSLATE} from "../injection-tokens/label-to-translate.injection-token";
import {DownloadToken} from "../models/download-token.model";
import {SolidifyHttpErrorResponseModel} from "../models/errors/solidify-http-error-response.model";
import {isTrue} from "../tools/is/is.tool";
import {SsrUtil} from "../utils/ssr.util";
import {UrlUtil} from "../utils/url.util";
import {AbstractBaseService} from "./abstract-base.service";
import {NotificationService} from "./notification.service";

@Injectable({
  providedIn: "root",
})
export class DownloadService extends AbstractBaseService {
  constructor(private readonly _httpClient: HttpClient,
              private readonly _apiService: ApiService,
              private readonly _notificationService: NotificationService,
              private readonly _cookieConsentService: CookieConsentService,
              @Inject(LABEL_TRANSLATE) private readonly _labelTranslate: LabelTranslateInterface,
              @Inject(ENVIRONMENT) private readonly _environment: DefaultSolidifyEnvironment) {
    super();
  }

  /**
   * Download using download token method if useDownloadToken is set to true. Otherwise download depending of isPublic attribute
   *
   * @param url
   * @param fileName
   * @param size
   * @param isPublic if true use download browser native. If false download in memory using token infos. Ignored if useDownloadToken is true
   */
  download(url: string, fileName: string, size?: number, isPublic: boolean = true): void {
    this._notificationService.showInformation(this._labelTranslate.coreFileDownloadStart);
    if (isTrue(this._environment.useDownloadToken)) {
      if (CookieConsentUtil.isFeatureAuthorized(CookieType.cookie, CookiePartialEnum.downloadToken)) {
        this._getAndStoreDownloadTokenThenDownload(url);
      } else {
        this._cookieConsentService.notifyFeatureDisabledBecauseCookieDeclined(CookieType.cookie, CookiePartialEnum.downloadToken);
        return;
      }
    } else {
      if (isPublic) {
        this.downloadBrowserNative(url);
      } else {
        this.subscribe(this.downloadInMemory(url, fileName, true));
      }
    }
  }

  downloadBrowserNative(url: string): void {
    SsrUtil.window?.open(url, "_self");
  }

  private _getAndStoreDownloadTokenThenDownload(url: string): void {
    this.subscribe(this._httpClient.get<DownloadToken>(this._generateDownloadTokenUrl(url), {
      withCredentials: true,
    }).pipe(
      take(1),
      tap(result => {
        this.downloadBrowserNative(url);
      }),
      catchError((error: SolidifyHttpErrorResponseModel) => {
        if (error.status !== HttpStatusCode.Unauthorized || error.error?.error !== OauthHelper.MFA_NEEDED) {
          this._notificationService.showError(this._labelTranslate.coreFileDownloadForbidden);
        }
        throw error;
      }),
    ));
  }

  private _generateDownloadTokenUrl(url: string): string {
    const endOfUrlExpected = SOLIDIFY_CONSTANTS.URL_SEPARATOR + ApiActionNamePartialEnum.DOWNLOAD;
    url = UrlUtil.removeParametersFromUrl(url);
    if (url.endsWith(endOfUrlExpected)) {
      return url.substring(0, url.length - endOfUrlExpected.length) + SOLIDIFY_CONSTANTS.URL_SEPARATOR + ApiActionNamePartialEnum.DOWNLOAD_TOKEN;
    }
    throw new SolidifyError(`Incompatible download url format for working with download token system. Disable useDownloadToken variable or change the format of download url to end by "${SOLIDIFY_CONSTANTS.URL_SEPARATOR + ApiActionNamePartialEnum.DOWNLOAD}"`);
  }

  // Prefer use download api to stream directly on disk
  downloadInMemory(url: string, fileName: string, withSaveAs: boolean = true): Observable<Blob> {
    let headers = new HttpHeaders();
    headers = headers.set("Content-Disposition", "attachment; filename=" + fileName);
    if (withSaveAs) {
      this._notificationService.showInformation(this._labelTranslate.coreFileDownloadStart);
    }
    return this._httpClient.get(url, {
      headers,
      responseType: "blob",
    }).pipe(
      tap((blobContent: Blob) => {
        if (withSaveAs) {
          this._notificationService.showSuccess(this._labelTranslate.coreFileDownloadDownloadedSuccessfully);
          saveAs(blobContent, fileName);
        }
      }),
      catchError((error: SolidifyHttpErrorResponseModel) => {
        if (withSaveAs) {
          this._notificationService.showError(this._labelTranslate.coreFileDownloadFail);
          if (error.status === HttpStatusCode.Forbidden) {
            this._notificationService.showError(this._labelTranslate.coreFileDownloadForbidden);
          } else {
            this._notificationService.showError(this._labelTranslate.coreFileDownloadFail);
          }
        }
        throw error;
      }),
    );
  }
}
